import asyncio
import base64
import logging
import secrets
from dataclasses import dataclass
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Any

from aiohttp import web
from cactus_schema.notification import CollectedHeader, CollectedNotification

from cactus_client_notifications.server.time import utc_now

logger = logging.getLogger(__name__)


class NotificationException(Exception):
    status_code: HTTPStatus  # What status code should be served by whatever is upstream of this request?

    def __init__(self, status_code: HTTPStatus, *args: Any) -> None:  # noqa: B042
        self.status_code = status_code
        super().__init__(*args)


def generate_unique_id() -> str:
    """Generates a URI safe string with a random value"""
    return base64.urlsafe_b64encode(secrets.token_bytes(24)).decode()


async def generate_collected_notification(request: web.Request) -> CollectedNotification:
    """Generates a CollectedNotification from an incoming web request"""

    headers = [CollectedHeader(name=name, value=val) for name, val in request.headers.items()]
    if request.body_exists:
        body = await request.text()
    else:
        body = ""

    return CollectedNotification(
        method=request.method, headers=headers, body=body, received_at=utc_now(), remote=request.remote
    )


class EndpointData:
    notifications: list[CollectedNotification]
    enabled: bool  # Is this endpoint enabled?
    created_at: datetime  # When was this endpoint created?
    interacted_at: datetime  # When did the last notification get received / endpoint get touched

    def __init__(self) -> None:
        self.notifications = []
        self.enabled = True
        self.interacted_at = utc_now()
        self.created_at = utc_now()


@dataclass
class EndpointMetadata:
    """Basic metadata about an Endpoint in the EndpointStore"""

    endpoint_id: str
    total_notifications: int
    enabled: bool  # Is this endpoint enabled?
    created_at: datetime  # When was this endpoint created?
    interacted_at: datetime  # When did the last notification get received / endpoint get touched


class EndpointStore:
    lock: asyncio.Lock
    max_active_endpoints: int
    max_endpoint_notifications: int

    _store: dict[str, EndpointData]

    def __init__(self, max_active_endpoints: int, max_endpoint_notifications: int):
        self.lock = asyncio.Lock()
        self.max_active_endpoints = max_active_endpoints
        self.max_endpoint_notifications = max_endpoint_notifications
        self._store = {}

    async def create_endpoint(self) -> str:
        """Creates a new endpoint with a unique ID. Returns that ID

        Can raise NotificationException"""
        async with self.lock:
            current_endpoint_count = len(self._store)
            if current_endpoint_count >= self.max_active_endpoints:
                raise NotificationException(
                    HTTPStatus.INSUFFICIENT_STORAGE,
                    f"There are already {current_endpoint_count} endpoints and max is {self.max_active_endpoints}",
                )

            new_id = generate_unique_id()
            if new_id in self._store:
                raise NotificationException(
                    HTTPStatus.INTERNAL_SERVER_ERROR, f"ID generation collision. '{new_id}' already exists"
                )

            logger.info(f"Created endpoint {new_id}")
            self._store[new_id] = EndpointData()
            return new_id

    async def update_endpoint(self, endpoint_id: str, enabled: bool) -> None:
        """Tries to update settings for endpoint_id. Raises NotificationException if this can't be done"""
        async with self.lock:
            data = self._store.get(endpoint_id, None)
            if data is None:
                raise NotificationException(
                    HTTPStatus.NOT_FOUND, f"No endpoint with ID {endpoint_id} exists (or it's been removed)."
                )

            data.enabled = enabled
            data.interacted_at = utc_now()
            logger.info(f"Updated endpoint {endpoint_id} - enabled={enabled}")

    async def try_delete_endpoint(self, endpoint_id: str) -> bool:
        """Tries to delete endpoint_id - returns True if it exists and was deleted. False if it DNE"""
        async with self.lock:
            exists = endpoint_id in self._store
            if exists:
                logger.info(f"Deleted endpoint {endpoint_id}")
                del self._store[endpoint_id]

            return exists

    async def add_notification(self, endpoint_id: str, notification: CollectedNotification) -> None:
        """Adds a notification to the specified endpoint_id - raises NotificationException if this can't be done."""
        async with self.lock:
            data = self._store.get(endpoint_id, None)
            if data is None:
                raise NotificationException(
                    HTTPStatus.NOT_FOUND, f"No endpoint with ID {endpoint_id} exists (or it's been removed)."
                )

            if not data.enabled:
                raise NotificationException(
                    HTTPStatus.INTERNAL_SERVER_ERROR,
                    f"Endpoint {endpoint_id} is temporarily disabled to simulate an outage.",
                )

            if len(data.notifications) >= self.max_endpoint_notifications:
                raise NotificationException(
                    HTTPStatus.INSUFFICIENT_STORAGE,
                    f"Endpoint {endpoint_id} has exceeded the max notifications ({self.max_endpoint_notifications})",
                )

            data.interacted_at = utc_now()
            data.notifications.append(notification)

            logger.info(
                f"Added {notification.method} notification ({len(notification.body)} bytes) to endpoint {endpoint_id}"
            )

    async def collect_notifications(self, endpoint_id: str) -> list[CollectedNotification]:
        """Collects all notifications for the specified endpoint_id. Notifications will be "consumed".

        Raises NotificationException on error."""
        async with self.lock:
            data = self._store.get(endpoint_id, None)
            if data is None:
                raise NotificationException(
                    HTTPStatus.NOT_FOUND, f"No endpoint with ID {endpoint_id} exists (or it's been removed)."
                )

            collected_notifications = data.notifications
            data.interacted_at = utc_now()
            data.notifications = []

            logger.info(f"Collected {len(collected_notifications)} notifications from endpoint {endpoint_id}")
            return collected_notifications

    async def cleanup_expired_endpoints(self, now: datetime, max_idle: timedelta, max_duration: timedelta) -> None:
        """Enumerates all endpoints, removing any that have reached their max duration/idle time"""
        async with self.lock:
            expired_endpoint_ids: list[str] = []
            for endpoint_id, data in self._store.items():
                current_duration = now - data.created_at
                if current_duration > max_duration:
                    logger.info(
                        f"Endpoint {endpoint_id} duration {current_duration} has exceeded max duration {max_duration}"
                    )
                    expired_endpoint_ids.append(endpoint_id)
                    continue

                current_idle = now - data.interacted_at
                if current_idle > max_idle:
                    logger.info(f"Endpoint {endpoint_id} idle time {current_idle} has exceeded max idle {max_idle}")
                    expired_endpoint_ids.append(endpoint_id)
                    continue

            for endpoint_id in expired_endpoint_ids:
                logger.info(f"Cleanup has deleted endpoint {endpoint_id}")
                del self._store[endpoint_id]

    async def get_endpoint_metadata(self) -> list[EndpointMetadata]:
        async with self.lock:
            return [
                EndpointMetadata(
                    endpoint_id,
                    len(endpoint.notifications),
                    endpoint.enabled,
                    endpoint.created_at,
                    endpoint.interacted_at,
                )
                for endpoint_id, endpoint in self._store.items()
            ]



import asyncio
import contextlib
import logging
import os
import sys
from datetime import timedelta
from typing import AsyncGenerator

from aiohttp import web
from cactus_schema.notification import uri

import cactus_client_notifications.server.shared as shared
from cactus_client_notifications.server import handler
from cactus_client_notifications.server.endpoint_store import EndpointStore
from cactus_client_notifications.server.settings import ServerSettings, ServerStats
from cactus_client_notifications.server.time import utc_now

logger = logging.getLogger(__name__)


async def periodic_task(app: web.Application) -> None:
    """Periodic task called app[APPKEY_PERIOD_SEC]

    Args:
        app (web.Application): The AIOHTTP application instance.
    """
    server_settings = app[shared.APPKEY_SERVER_SETTINGS]
    store = app[shared.APPKEY_NOTIFICATION_STORE]

    while True:

        # Sleep first - we don't need to initiate a cleanup immediately
        await asyncio.sleep(server_settings.cleanup_frequency.total_seconds())

        try:
            await store.cleanup_expired_endpoints(
                utc_now(), server_settings.max_endpoint_idle_duration, server_settings.max_endpoint_duration
            )
        except Exception as exc:
            # Catch and log uncaught exceptions to prevent periodic task from hanging
            logger.error("Uncaught exception in periodic task", exc_info=exc)


async def setup_periodic_task(app: web.Application) -> AsyncGenerator:
    """Setup periodic task.

    The periodic task is accessible through app[APPKEY_PERIODIC_TASKS].
    The code for the task is defined in the function 'periodic_task'.
    """
    app[shared.APPKEY_PERIODIC_TASK] = asyncio.create_task(periodic_task(app))

    yield

    app[shared.APPKEY_PERIODIC_TASK].cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await app[shared.APPKEY_PERIODIC_TASK]


def create_app() -> web.Application:

    app = web.Application()

    # Check env for basic settings
    ENV_APP_PORT = int(os.getenv("APP_PORT", 8080))
    ENV_PUBLIC_SERVER_URL = os.getenv("SERVER_URL", f"http://localhost:{ENV_APP_PORT}")
    ENV_MOUNT_POINT = os.getenv("MOUNT_POINT", "/")
    ENV_MAX_IDLE_DURATION_SECONDS = float(os.getenv("MAX_IDLE_DURATION_SECONDS", 3600))
    ENV_MAX_DURATION_SECONDS = float(os.getenv("MAX_DURATION_SECONDS", (3600 * 24 * 3) + 3600))
    ENV_MAX_ACTIVE_ENDPOINTS = int(os.getenv("MAX_ACTIVE_ENDPOINTS", 1024))
    ENV_MAX_ENDPOINT_NOTIFICATIONS = int(os.getenv("MAX_ENDPOINT_NOTIFICATIONS", 100))
    ENV_CLEANUP_FREQUENCY_SECONDS = float(os.getenv("CLEANUP_FREQUENCY_SECONDS", 120))
    server_settings = ServerSettings(
        port=ENV_APP_PORT,
        public_server_url=ENV_PUBLIC_SERVER_URL,
        mount_point=ENV_MOUNT_POINT,
        max_endpoint_idle_duration=timedelta(seconds=ENV_MAX_IDLE_DURATION_SECONDS),
        max_endpoint_duration=timedelta(seconds=ENV_MAX_DURATION_SECONDS),
        cleanup_frequency=timedelta(seconds=ENV_CLEANUP_FREQUENCY_SECONDS),
        started_at=utc_now(),
        max_active_endpoints=ENV_MAX_ACTIVE_ENDPOINTS,
        max_endpoint_notifications=ENV_MAX_ENDPOINT_NOTIFICATIONS,
    )

    app[shared.APPKEY_NOTIFICATION_STORE] = EndpointStore(
        max_endpoint_notifications=server_settings.max_endpoint_notifications,
        max_active_endpoints=server_settings.max_active_endpoints,
    )
    app[shared.APPKEY_SERVER_SETTINGS] = server_settings
    app[shared.APPKEY_SERVER_STATS] = ServerStats()

    # Add routes for Test Runner
    mount = server_settings.mount_point
    app.router.add_route(
        "POST", handler.path_join(mount, uri.URI_MANAGE_ENDPOINT_LIST), handler.post_manage_endpoint_list
    )
    app.router.add_route("GET", handler.path_join(mount, uri.URI_MANAGE_ENDPOINT), handler.get_manage_endpoint)
    app.router.add_route("PUT", handler.path_join(mount, uri.URI_MANAGE_ENDPOINT), handler.put_manage_endpoint)
    app.router.add_route("DELETE", handler.path_join(mount, uri.URI_MANAGE_ENDPOINT), handler.delete_manage_endpoint)
    app.router.add_route("*", handler.path_join(mount, uri.URI_ENDPOINT), handler.webhook_endpoint)
    app.router.add_route("GET", handler.path_join(mount, uri.URI_MANAGE_SERVER), handler.get_manage_server)

    # Start the periodic task
    app.cleanup_ctx.append(setup_periodic_task)

    return app


def create_app_with_logging() -> web.Application:
    ENV_LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO")

    logging.basicConfig(
        level=ENV_LOG_LEVEL,
        format="%(asctime)s.%(msecs)03d [%(levelname)s] [%(name)s] (%(funcName)s): %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    app = create_app()

    return app


app = create_app_with_logging()
if __name__ == "__main__":
    web.run_app(app, port=app[shared.APPKEY_SERVER_SETTINGS].port)



import http
import logging
from importlib.metadata import version

from aiohttp import ContentTypeError, web
from cactus_schema.notification import (
    CollectEndpointResponse,
    ConfigureEndpointRequest,
    CreateEndpointResponse,
    uri,
)

from cactus_client_notifications.server.endpoint_store import (
    NotificationException,
    generate_collected_notification,
)
from cactus_client_notifications.server.settings import ServerSettings
from cactus_client_notifications.server.shared import (
    APPKEY_NOTIFICATION_STORE,
    APPKEY_SERVER_SETTINGS,
    APPKEY_SERVER_STATS,
)

VERSION = version("cactus_client_notifications")

logger = logging.getLogger(__name__)


def path_join(*parts: str) -> str:
    joinable_parts: list[str] = []

    last: str | None = None
    for next in parts:
        next = next.strip()
        if not next:
            continue

        if last is None:
            joinable_parts.append(next)
            last = next
            continue

        # Here is where the join logic happens
        if last.endswith("/"):
            if next.startswith("/"):
                if len(next) == 1:
                    continue  # Don't add an empty string (once we strip the /)
                joinable_parts.append(next[1:])
            else:
                joinable_parts.append(next)
        else:
            if next.startswith("/"):
                joinable_parts.append(next)
            else:
                joinable_parts.append("/")
                joinable_parts.append(next)
        last = next

    return "".join(joinable_parts)


def generate_public_uri(server_settings: ServerSettings, endpoint_id: str) -> str:
    """Generates the public facing URI for a specific endpoint_id"""
    return path_join(
        server_settings.public_server_url, server_settings.mount_point, uri.URI_ENDPOINT.format(endpoint_id=endpoint_id)
    )


async def post_manage_endpoint_list(request: web.Request) -> web.Response:
    """Expects an empty POST body. Creates a new endpoint

    Args:
        request: An aiohttp.web.Request instance.

    Returns:
        aiohttp.web.Response: Encodes a CreateEndpointResponse on success

        a 201 (CREATED) on success
        a 507 (INSUFFICIENT_STORAGE) if the webserver has too many notification endpoints at this moment
    """

    request.app[APPKEY_SERVER_STATS].total_created_webhooks += 1
    logger.info(f"Creating endpoint for {request.remote}")

    try:
        endpoint_id = await request.app[APPKEY_NOTIFICATION_STORE].create_endpoint()
    except NotificationException as exc:
        request.app[APPKEY_SERVER_STATS].total_created_webhook_errors += 1
        logger.error("Error creating endpoint", exc_info=exc)
        return web.Response(status=exc.status_code, text=str(exc))

    create_response = CreateEndpointResponse(
        endpoint_id=endpoint_id,
        fully_qualified_endpoint=generate_public_uri(request.app[APPKEY_SERVER_SETTINGS], endpoint_id),
    )

    return web.Response(status=http.HTTPStatus.CREATED, content_type="application/json", text=create_response.to_json())


async def get_manage_endpoint(request: web.Request) -> web.Response:
    """Performs a collection of notifications for the requested endpoint id. This will "consume" all notifications
    that are collected.

    Args:
        request: An aiohttp.web.Request instance.

    Returns:
        aiohttp.web.Response: Encodes a CollectEndpointResponse on success

        a 200 (OK) on success - yielding a CollectEndpointResponse as JSON
        a 404 (NOT_FOUND) if the endpoint has been deleted or the endpoint_id is invalid
    """

    endpoint_id = request.match_info.get("endpoint_id")
    if not endpoint_id:
        return web.Response(status=http.HTTPStatus.BAD_REQUEST, text="endpoint_id couldn't be extracted from the path.")

    request.app[APPKEY_SERVER_STATS].total_collections += 1
    logger.info(f"Collecting endpoint {endpoint_id} for {request.remote}")

    try:
        collected_notifications = await request.app[APPKEY_NOTIFICATION_STORE].collect_notifications(endpoint_id)
    except NotificationException as exc:
        request.app[APPKEY_SERVER_STATS].total_collection_errors += 1
        logger.error(f"Error updating config for {endpoint_id}", exc_info=exc)
        return web.Response(status=exc.status_code, text=str(exc))

    collect_response = CollectEndpointResponse(
        notifications=collected_notifications,
    )

    return web.Response(status=http.HTTPStatus.OK, content_type="application/json", text=collect_response.to_json())


async def put_manage_endpoint(request: web.Request) -> web.Response:
    """Updates the settings for an existing endpoint. Expects a ConfigureEndpointRequest in the request body

    Args:
        request: An aiohttp.web.Request instance.

    Returns:
        aiohttp.web.Response: Encodes a CreateEndpointResponse on success

        a 204 (NO_CONTENT) on success.
        a 404 (NOT_FOUND) if the endpoint has been deleted or the endpoint_id is invalid
    """

    endpoint_id = request.match_info.get("endpoint_id")
    if not endpoint_id:
        return web.Response(status=http.HTTPStatus.BAD_REQUEST, text="endpoint_id couldn't be extracted from the path.")

    request.app[APPKEY_SERVER_STATS].total_configures += 1

    try:
        raw_json = await request.text()
    except ContentTypeError:
        request.app[APPKEY_SERVER_STATS].total_configure_errors += 1
        return web.Response(status=http.HTTPStatus.BAD_REQUEST, text="Missing JSON body")

    configure_request = ConfigureEndpointRequest.from_json(raw_json)
    if isinstance(configure_request, list):
        request.app[APPKEY_SERVER_STATS].total_configure_errors += 1
        return web.Response(status=http.HTTPStatus.BAD_REQUEST, text="Singular ConfigureEndpointRequest is required.")

    logger.info(f"Configuring endpoint {endpoint_id} with {configure_request} for {request.remote}")

    try:
        await request.app[APPKEY_NOTIFICATION_STORE].update_endpoint(endpoint_id, enabled=configure_request.enabled)
    except NotificationException as exc:
        request.app[APPKEY_SERVER_STATS].total_configure_errors += 1
        logger.error(f"Error configuring {endpoint_id} with {configure_request}", exc_info=exc)
        return web.Response(status=exc.status_code, text=str(exc))

    return web.Response(status=http.HTTPStatus.NO_CONTENT)


async def delete_manage_endpoint(request: web.Request) -> web.Response:
    """Deletes an existing endpoint based on the endpoint_id in the path. All uncollected notifications will be lost.

    Args:
        request: An aiohttp.web.Request instance.

    Returns:
        aiohttp.web.Response: Encodes a CreateEndpointResponse on success

        a 204 (NO_CONTENT) on success.
        a 404 (NOT_FOUND) if the endpoint has been deleted or the endpoint_id is invalid
    """

    endpoint_id = request.match_info.get("endpoint_id")
    if not endpoint_id:
        return web.Response(status=http.HTTPStatus.BAD_REQUEST, text="endpoint_id couldn't be extracted from the path.")

    request.app[APPKEY_SERVER_STATS].total_deletes += 1

    logger.info(f"Deleting endpoint {endpoint_id} for {request.remote}")

    try:
        await request.app[APPKEY_NOTIFICATION_STORE].try_delete_endpoint(endpoint_id)
    except NotificationException as exc:
        request.app[APPKEY_SERVER_STATS].total_delete_errors += 1
        logger.error(f"Error deleting {endpoint_id}", exc_info=exc)
        return web.Response(status=exc.status_code, text=str(exc))

    return web.Response(status=http.HTTPStatus.NO_CONTENT)


async def webhook_endpoint(request: web.Request) -> web.Response:
    """This is the endpoint that will handle ALL incoming 2030.5 notifications from the utility server. It will try
    to report success for everything and log the contents of the incoming request.

    Args:
        request: An aiohttp.web.Request instance.

    Returns:
        aiohttp.web.Response: Encodes a CreateEndpointResponse on success

        a 200 (OK) on success.
        a 404 (NOT_FOUND) if the endpoint has been deleted or the endpoint_id is invalid
        a 500 (INTERNAL_SERVER_ERROR) if the endpoint has been disabled
        a 507 (INSUFFICIENT_STORAGE) if the endpoint has too many uncollected notifications
    """

    endpoint_id = request.match_info.get("endpoint_id")
    if not endpoint_id:
        return web.Response(status=http.HTTPStatus.NOT_FOUND)

    request.app[APPKEY_SERVER_STATS].total_notifications += 1

    try:
        collected_notification = await generate_collected_notification(request)
    except Exception as exc:
        request.app[APPKEY_SERVER_STATS].total_notification_errors += 1
        logger.error(f"Error parsing incoming webhook request for {endpoint_id} from {request.remote}", exc_info=exc)
        return web.Response(status=http.HTTPStatus.BAD_REQUEST)

    logger.info(
        f"{collected_notification.method} notification ({len(collected_notification.body)}) at {endpoint_id}"
        + f"from {request.remote}."
    )

    try:
        await request.app[APPKEY_NOTIFICATION_STORE].add_notification(endpoint_id, collected_notification)
    except NotificationException as exc:
        logger.error(f"Error adding notification to {endpoint_id}", exc_info=exc)
        return web.Response(status=exc.status_code, text=str(exc))

    return web.Response(status=http.HTTPStatus.OK)


async def get_manage_server(request: web.Request) -> web.Response:
    """Returns a text/plain readout of the current server status

    Returns:
        aiohttp.web.Response: Encodes a plaintext response body

        a 200 (OK) on success.
    """

    endpoint_metadata = await request.app[APPKEY_NOTIFICATION_STORE].get_endpoint_metadata()
    stats = request.app[APPKEY_SERVER_STATS]
    settings = request.app[APPKEY_SERVER_SETTINGS]

    sep = "-" * 40
    lines: list[str] = [
        f"CACTUS Client Notifications Server {VERSION}",
        sep,
        "SETTINGS",
        f"Started: {stats.created_at.isoformat()}",
        f"Max Active Endpoints: {settings.max_active_endpoints}",
        f"Max Endpoint Notifications: {settings.max_endpoint_notifications}",
        f"Cleanup Frequency: {settings.cleanup_frequency.total_seconds()}s",
        f"Max Endpoint Idle Time: {settings.max_endpoint_idle_duration.total_seconds()}s",
        f"Max Endpoint Total Time: {settings.max_endpoint_duration.total_seconds()}s",
        f"Mount Point: {settings.mount_point}",
        f"Public URL: {settings.public_server_url}",
        sep,
        "STATS",
        f"Total created notifications: {stats.total_notifications} (with {stats.total_notification_errors} errors)",
        f"Total collection requests: {stats.total_collections} (with {stats.total_collection_errors} errors)",
        f"Total received notifications: {stats.total_notifications} (with {stats.total_notification_errors} errors)",
        f"Total configure requests: {stats.total_configures} (with {stats.total_configure_errors} errors)",
        sep,
        f"ENDPOINTS ({len(endpoint_metadata)} total)",
    ]

    for endpoint in endpoint_metadata:
        lines.append("")
        lines.append(f"Endpoint {endpoint.endpoint_id[:4]}...")  # Don't show the full ID - just enough to watch things
        lines.append(f"    {endpoint.total_notifications} uncollected notifications")
        lines.append(f"    Enabled: {endpoint.enabled}")
        lines.append(f"    Created: {endpoint.created_at.isoformat()}")
        lines.append(f"    Last Touched: {endpoint.interacted_at.isoformat()}")

    lines.append(sep)

    return web.Response(status=http.HTTPStatus.OK, text="\n".join(lines))



from datetime import datetime, timedelta, timezone


def utc_now() -> datetime:
    """Returns an unambiguous timezone aware (UTC) datetime representing this moment"""
    return datetime.now(tz=timezone.utc)


def relative_time(delta: timedelta) -> str:
    """Returns a human readable string representing delta"""

    total_seconds = delta.total_seconds()
    if total_seconds >= 0:
        sign = "+"
    else:
        sign = "-"

    magnitude = abs(total_seconds)
    if magnitude < 5:
        return f"{sign}{int(magnitude*1000)}ms"
    elif magnitude < 120:
        return f"{sign}{magnitude:.1f}s"
    else:
        return f"{sign}{int(magnitude) // 60}m{int(magnitude) % 60}s"



import asyncio

from aiohttp import web

from cactus_client_notifications.server.endpoint_store import EndpointStore
from cactus_client_notifications.server.settings import ServerSettings, ServerStats

APPKEY_NOTIFICATION_STORE = web.AppKey("notification-store", EndpointStore)
APPKEY_SERVER_SETTINGS = web.AppKey("server-settings", ServerSettings)
APPKEY_PERIODIC_TASK = web.AppKey("periodic-task", asyncio.Task)
APPKEY_SERVER_STATS = web.AppKey("server-stats", ServerStats)



from dataclasses import dataclass, field
from datetime import datetime, timedelta

from cactus_client_notifications.server.time import utc_now


@dataclass(frozen=True)
class ServerSettings:
    port: int  # The app port that is being listened on
    public_server_url: str  # The public facing URL that all endpoints will be exposed under
    mount_point: str  # The path prefix that all served endpoints (and API endpoint endpoints) will be exposed under

    max_endpoint_idle_duration: (
        timedelta  # How long can an endpoint go without receiving a request before being removed
    )
    max_endpoint_duration: timedelta  # How long can an endpoint go for without being shutdown
    cleanup_frequency: timedelta  # How often to check for expired endpoints
    started_at: datetime

    max_active_endpoints: int  # How many endpoints can be managed at the same time
    max_endpoint_notifications: int  # How many notifications can an endpoint cache before dropping requests


@dataclass
class ServerStats:
    created_at: datetime = field(default_factory=utc_now)
    total_notifications: int = 0  # How many notifications have landed at the webhook since the server started
    total_notification_errors: int = 0  # How many exceptions caused by creating notification
    total_created_webhooks: int = 0  # How many webhooks have been created
    total_created_webhook_errors: int = 0  # How many exception caused by incoming webhook notifications
    total_collections: int = 0  # How many times has collect been called
    total_collection_errors: int = 0  # How many collect errors have been raised
    total_deletes: int = 0  # How many times has delete been called
    total_delete_errors: int = 0  # How many delete errors have been raised
    total_configures: int = 0  # How many times has the configure endpoint been called
    total_configure_errors: int = 0  # How many configure errors have been raised
